@@ -1,5 +1,5 @@
 /* $Copyright: $
- * Copyright (c) 1996 by Steve Baker (ice@mama.indstate.edu)
+ * Copyright (c) 1996,1997 by Steve Baker (ice@mama.indstate.edu)
  * All Rights reserved
  *
  * This software is provided as is without any express or implied
@@ -7,6 +7,7 @@
  * of merchantability and fitness for a particular purpose.
  */
 #include <stdio.h>
+#include <time.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -15,45 +16,63 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <linux/limits.h>
+#include <pwd.h>
+#include <grp.h>
 
-static char *version = "$Version: $ tree v1.1 (c) 1996 by Steve Baker, Thomas Moore $";
+static char *version = "$Version: $ tree v1.2 (c) 1996,1997 by Steve Baker, Thomas Moore $";
 
 void *strcpy(), *malloc();
 char **parse(), **split(), *getenv();
 
-#define SCOPY(x)	strcpy(malloc(strlen(x)+1),(x))
+#define scopy(x)	strcpy(malloc(strlen(x)+1),(x))
 
 #define MAXDIRLEVEL	PATH_MAX/2
 #define MINIT		30  /* number of dir entries to initially allocate */
 #define MINC		20  /* allocation increment */
 
 enum {FALSE=0, TRUE};
+/* sort type */
+enum {ALPHA, TIME};
 
-u_char dflag, lflag, pflag, sflag, Fflag, aflag, fflag;
+u_char dflag, lflag, pflag, sflag, Fflag, aflag, fflag, uflag, gflag;
+u_char qflag, Nflag, Dflag;
+u_char sort = ALPHA;
 u_char noindent, force_color, nocolor, xdev;
-char *pattern = NULL;
+char *pattern = NULL, *ipattern = NULL;
 
 int cmp();
 struct _info **read_dir();
-char *prot();
+char *prot(), *uidtoname(), *gidtoname(), *do_date();
 
 int dirs[MAXDIRLEVEL];
 struct _info {
   char *name;
   char *lnk;
-  u_char isdir : 1;
-  u_char issok : 1;
-  u_char isexe : 1;
+  u_char isdir  : 1;
+  u_char issok  : 1;
+  u_char isexe  : 1;
+  u_char isfifo : 1;
+  u_char orphan : 1;
   u_short mode, lnkmode, uid, gid;
   u_long size;
+  time_t atime, ctime, mtime;
   dev_t dev;
 };
 
+/* Faster uid/gid -> name lookup with hash(tm)(r)(c) tables! */
+#define HASH(x)		((x)&255)
+struct xtable {
+  u_short xid;
+  char *name;
+  struct xtable *nxt;
+} *gtable[256], *utable[256];
+
 /* Hacked in DIR_COLORS support ---------------------------------------------- */
 
 enum {
   CMD_COLOR, CMD_OPTIONS, CMD_TERM, CMD_EIGHTBIT, COL_NORMAL, COL_FILE, COL_DIR,
-  COL_LINK, COL_FIFO, COL_SOCK, COL_BLK, COL_CHR, COL_EXEC, DOT_EXTENSION, ERROR
+  COL_LINK, COL_FIFO, COL_SOCK, COL_BLK, COL_CHR, COL_EXEC, DOT_EXTENSION, ERROR,
+  COL_ORPHAN, COL_MISSING, COL_LEFTCODE, COL_RIGHTCODE, COL_ENDCODE
 };
 
 char colorize = FALSE, ansilines = FALSE;
@@ -61,7 +80,8 @@ char *term, termmatch = FALSE, istty;
 
 char *norm_flgs = NULL, *file_flgs = NULL, *dir_flgs = NULL, *link_flgs = NULL;
 char *fifo_flgs = NULL, *sock_flgs = NULL, *block_flgs = NULL, *char_flgs = NULL;
-char *exec_flgs = NULL;
+char *exec_flgs = NULL, *orphan_flgs = NULL, *missing_flgs = NULL, *leftcode = NULL;
+char *rightcode = NULL, *endcode = NULL;
 
 struct extensions {
   char *ext;
@@ -75,11 +95,12 @@ int argc;
 char **argv;
 {
   char **dirname = NULL;
-  int i,j,n,p,dtotal,ftotal,colored;
+  int i,j,n,p,q,dtotal,ftotal,colored = FALSE;
   struct stat st;
 
-  n = p = dtotal = ftotal = 0;
-  aflag = dflag = fflag = lflag = pflag = sflag = Fflag = FALSE;
+  q = p = dtotal = ftotal = 0;
+  aflag = dflag = fflag = lflag = pflag = sflag = Fflag = uflag = gflag = FALSE;
+  Dflag = qflag = Nflag = FALSE;
   noindent = force_color = nocolor = xdev = FALSE;
 
   for(n=i=1;i<argc;i=n) {
@@ -87,6 +108,12 @@ char **argv;
     if (argv[i][0] == '-') {
       for(j=1;argv[i][j];j++) {
 	switch(argv[i][j]) {
+	  case 'N':
+	    Nflag = TRUE;
+	    break;
+	  case 'q':
+	    qflag = TRUE;
+	    break;
 	  case 'd':
 	    dflag = TRUE;
 	    break;
@@ -96,6 +123,12 @@ char **argv;
 	  case 's':
 	    sflag = TRUE;
 	    break;
+	  case 'u':
+	    uflag = TRUE;
+	    break;
+	  case 'g':
+	    gflag = TRUE;
+	    break;
 	  case 'f':
 	    fflag = TRUE;
 	    break;
@@ -123,19 +156,28 @@ char **argv;
 	  case 'P':
 	    pattern = argv[n++];
 	    break;
+	  case 'I':
+	    ipattern = argv[n++];
+	    break;
 	  case 'A':
 	    ansilines = TRUE;
 	    break;
+	  case 'D':
+	    Dflag = TRUE;
+	    break;
+	  case 't':
+	    sort = TIME;
+	    break;
 	  default:
-	    fprintf(stderr,"tree: Invalid argument.\nusage: tree [-adflpsFiCn] [<directory list>]\n");
+	    fprintf(stderr,"tree: Invalid argument -`%c'.\nusage: tree [-adfgilnpqstuxACDFN] [-P pattern] [-I pattern] [<directory list>]\n",argv[i][j]);
 	    exit(1);
 	    break;
 	}
       }
     } else {
-      if (!dirname) dirname = (char **)malloc(sizeof(char *) * (n=MINIT));
-      else if (p == (n-1)) dirname = (char **)realloc(dirname,sizeof(char *) * (n+=MINC));
-      dirname[p++] = SCOPY(argv[i]);
+      if (!dirname) dirname = (char **)malloc(sizeof(char *) * (q=MINIT));
+      else if (p == (q-1)) dirname = (char **)realloc(dirname,sizeof(char *) * (q+=MINC));
+      dirname[p++] = scopy(argv[i]);
     }
   }
   if (p) dirname[p] = NULL;
@@ -144,17 +186,17 @@ char **argv;
 
   if (dirname) {
     for(colored=i=0;dirname[i];i++,colored=0) {
-      if (colorize && stat(dirname[i],&st) >= 0)
-	colored = color(st.st_mode,dirname[i]);
-      printf("%s",dirname[i]);
-      if (colored) printf("\033[%sm",norm_flgs);
+      if (colorize && (n=lstat(dirname[i],&st)) >= 0)
+	colored = color(st.st_mode,dirname[i],n<0,FALSE);
+      printit(dirname[i]);
+      if (colored) printf("%s",endcode);
       listdir(dirname[i],&dtotal,&ftotal,0,0);
     }
   } else {
-    if (colorize && stat(".",&st) >= 0)
-      colored = color(st.st_mode,".");
+    if (colorize && (n=lstat(".",&st)) >= 0)
+      colored = color(st.st_mode,".",n<0,FALSE);
     putchar('.');
-    if (colored) printf("\033[%sm",norm_flgs);
+    if (colored) printf("%s",endcode);
     listdir(".",&dtotal,&ftotal,0,0);
   }
   if (dflag)
@@ -173,7 +215,7 @@ dev_t dev;
   char path[PATH_MAX+1], nlf = FALSE, colored = FALSE;
   struct _info **dir, **sav;
   struct stat sb;
-  int n;
+  int n,x, m,e;
 
   if (xdev && lev == 0) {
     stat(d,&sb);
@@ -196,30 +238,46 @@ dev_t dev;
   while(*dir) {
     if (!noindent) indent();
 
-    if (pflag || sflag) printf("[");
+    if (pflag || sflag || uflag || gflag || Dflag) printf("[");
     if (pflag) printf("%s",prot((*dir)->mode));
-    if (pflag && sflag) printf(" ");
+    if (pflag && (uflag || gflag || sflag || Dflag)) printf(" ");
+    if (uflag) printf("%-8.8s",uidtoname((*dir)->uid));
+    if (uflag && (gflag || sflag || Dflag)) printf(" ");
+    if (gflag) printf("%-8.8s",gidtoname((*dir)->gid));
+    if (gflag && (sflag || Dflag)) printf(" ");
     if (sflag) printf("%9d",(*dir)->size);
-    if (pflag || sflag) printf("]  ");
+    if (sflag && Dflag) printf(" ");
+    if (Dflag) printf("%s",do_date((*dir)->mtime));
+    if (pflag || sflag || uflag || gflag || Dflag) printf("]  ");
 
     if (colorize)
-      colored = color((*dir)->mode,(*dir)->name);
+      colored = color((*dir)->mode,(*dir)->name,(*dir)->orphan,FALSE);
 
-    if (fflag) printf("%s/%s",d,(*dir)->name);
-    else printf("%s",(*dir)->name);
+    if (fflag) sprintf(path,"%s/%s",d,(*dir)->name);
+    else sprintf(path,"%s",(*dir)->name);
+    printit(path);
 
-    if (colored) printf("\033[%sm",norm_flgs);
+    if (colored) printf("%s",endcode);
     
-    if (Fflag)
+    if (Fflag && !(*dir)->lnk)
       if (!dflag && (*dir)->isdir) putchar('/');
-      else if (!(*dir)->isdir && (*dir)->isexe) putchar('*');
       else if ((*dir)->issok) putchar('=');
+      else if ((*dir)->isfifo) putchar('|');
+      else if (!(*dir)->isdir && (*dir)->isexe) putchar('*');
 
     if ((*dir)->lnk) {
       printf(" -> ");
-      if (colorize) colored = color((*dir)->lnkmode,(*dir)->lnk);
-      printf("%s",(*dir)->lnk);
-      if (colored) printf("\033[%sm",norm_flgs);
+      if (colorize) colored = color((*dir)->lnkmode,(*dir)->lnk,(*dir)->orphan,TRUE);
+      printit((*dir)->lnk);
+      if (colored) printf("%s",endcode);
+      if (Fflag) {
+	m = (*dir)->lnkmode & S_IFMT;
+	e = ((*dir)->lnkmode & S_IEXEC) | ((*dir)->lnkmode & (S_IEXEC>>3)) | ((*dir)->lnkmode & (S_IEXEC>>6));
+	if (!dflag && m == S_IFDIR) putchar('/');
+	else if (m == S_IFSOCK) putchar('=');
+	else if (m == S_IFIFO) putchar('|');
+	else if (!(m == S_IFDIR) && e) putchar('*');
+      }
     }
 
     if ((*dir)->isdir) {
@@ -260,7 +318,7 @@ int *n;
   struct direct *ent;
   struct stat lst,st;
   DIR *d;
-  int ne, p = 0, len;
+  int ne, p = 0, len, rs;
 
   *n = 1;
   if ((d=opendir(dir)) == NULL) return NULL;
@@ -272,51 +330,51 @@ int *n;
     if (!aflag && ent->d_name[0] == '.') continue;
 
     sprintf(path,"%s/%s",dir,ent->d_name);
-    if (stat(path,&st) < 0) continue;
-    if (dflag && ((st.st_mode & S_IFMT) != S_IFDIR)) continue;
-
-    if ((st.st_mode & S_IFMT) != S_IFDIR && pattern && patmatch(ent->d_name,pattern) != 1) continue;
-
     if (lstat(path,&lst) < 0) continue;
-    if (pattern && ((lst.st_mode & S_IFMT) == S_IFLNK) && !lflag) continue;
+    if ((rs = stat(path,&st)) < 0) st.st_mode = 0;
+
+    if ((lst.st_mode & S_IFMT) != S_IFDIR && !(((st.st_mode & S_IFMT) == S_IFLNK) && lflag)) {
+      if (pattern && patmatch(ent->d_name,pattern) != 1) continue;
+      if (ipattern && patmatch(ent->d_name,ipattern) == 1) continue;
+    }
+
+    if (dflag && ((st.st_mode & S_IFMT) != S_IFDIR)) continue;
+    if (pattern && ((st.st_mode & S_IFMT) == S_IFLNK) && !lflag) continue;
+
 
     if (p == (ne-1)) dl = (struct _info **)realloc(dl,sizeof(struct _info) * (ne += MINC));
     dl[p] = (struct _info *)malloc(sizeof(struct _info));
 
+    dl[p]->name = scopy(ent->d_name);
+    dl[p]->mode = lst.st_mode;
+    dl[p]->uid = lst.st_uid;
+    dl[p]->gid = lst.st_gid;
+    dl[p]->size = lst.st_size;
+    dl[p]->dev = lst.st_dev;
+    dl[p]->lnk = NULL;
+    dl[p]->orphan = FALSE;
+
+    dl[p]->atime = lst.st_atime;
+    dl[p]->ctime = lst.st_ctime;
+    dl[p]->mtime = lst.st_mtime;
+
     if ((lst.st_mode & S_IFMT) == S_IFLNK) {
       if ((len=readlink(path,lbuf,PATH_MAX)) < 0) {
-	dl[p]->lnk = SCOPY("[Error reading symbolic link information]");
-	dl[p]->name = SCOPY(ent->d_name);
+	dl[p]->lnk = scopy("[Error reading symbolic link information]");
 	dl[p]->isdir = dl[p]->issok = dl[p]->isexe = FALSE;
-	dl[p]->mode = lst.st_mode;
-	dl[p]->lnkmode = st.st_mode;
-	dl[p]->uid = lst.st_uid;
-	dl[p]->gid = lst.st_gid;
-	dl[p]->dev = lst.st_dev;
-	dl[p++]->size = lst.st_size;
+	dl[p++]->lnkmode = st.st_mode;
 	continue;
       } else {
 	lbuf[len] = 0;
-	dl[p]->lnk = SCOPY(lbuf);
-	dl[p]->mode = lst.st_mode;
+	dl[p]->lnk = scopy(lbuf);
+	if (rs < 0) dl[p]->orphan = TRUE;
 	dl[p]->lnkmode = st.st_mode;
-	dl[p]->uid = lst.st_uid;
-	dl[p]->gid = lst.st_gid;
-	dl[p]->size = lst.st_size;
-	dl[p]->dev = lst.st_dev;
       }
-    } else {
-      dl[p]->lnk = NULL;
-      dl[p]->mode = st.st_mode;
-      dl[p]->uid = st.st_uid;
-      dl[p]->gid = st.st_gid;
-      dl[p]->size = st.st_size;
-      dl[p]->dev = st.st_dev;
     }
     
-    dl[p]->name = SCOPY(ent->d_name);
     dl[p]->isdir = ((st.st_mode & S_IFMT) == S_IFDIR);
     dl[p]->issok = ((st.st_mode & S_IFMT) == S_IFSOCK);
+    dl[p]->isfifo = ((st.st_mode & S_IFMT) == S_IFIFO);
     dl[p++]->isexe = (st.st_mode & S_IEXEC) | (st.st_mode & (S_IEXEC>>3)) | (st.st_mode & (S_IEXEC>>6));
   }
   closedir(d);
@@ -328,7 +386,10 @@ int *n;
 cmp(a,b)
 struct _info **a, **b;
 {
-  return strcmp((*a)->name,(*b)->name);
+  if (sort == ALPHA)
+    return strcmp((*a)->name,(*b)->name);
+  else
+    return (*a)->mtime < (*b)->mtime;
 }
 
 /*
@@ -400,7 +461,6 @@ char *buf,*pat;
  * They cried out for ANSI-lines (not really), but here they are, as an option
  * for the xterm and console capable among you, as a run-time option.
  */
-
 indent()
 {
   int i;
@@ -473,6 +533,102 @@ u_short m;
   return buf;
 }
 
+static char *month[] = {
+  "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
+};
+
+#define SIXMONTHS (6*31*24*60*60)
+
+char *do_date(t)
+time_t t;
+{
+  time_t c = time(0);
+  struct tm *tm;
+  static char buf[30];
+
+  if (t > c) return " The Future ";
+
+  tm = localtime(&t);
+
+  sprintf(buf,"%s %2d",month[tm->tm_mon],tm->tm_mday);
+  if (t+SIXMONTHS < c) sprintf(buf+6,"  %4d",1900+tm->tm_year);
+  else sprintf(buf+6," %2d:%02d",tm->tm_hour,tm->tm_min);
+  return buf;
+}
+
+char *uidtoname(uid)
+{
+  struct xtable *o, *p, *t;
+  struct passwd *ent;
+  char ubuf[6];
+  int uent = HASH(uid);
+
+  for(o = p = utable[uent]; p ; p=p->nxt) {
+    if (uid == p->xid) return p->name;
+    else if (uid < p->xid) break;
+    o = p;
+  }
+  /* uh oh, time to do a real lookup */
+  t = malloc(sizeof(struct xtable));
+  if ((ent = getpwuid(uid)) != NULL) t->name = scopy(ent->pw_name);
+  else {
+    sprintf(ubuf,"%d",uid);
+    t->name = scopy(ubuf);
+  }
+  t->xid = uid;
+  t->nxt = p;
+  if (o == utable[uent]) utable[uent] = t;
+  else o->nxt = t;
+  return t->name;
+}
+
+char *gidtoname(gid)
+{
+  struct xtable *o, *p, *t;
+  struct group *ent;
+  char gbuf[6];
+  int gent = HASH(gid);
+
+  for(o = p = gtable[gent]; p ; p=p->nxt) {
+    if (gid == p->xid) return p->name;
+    else if (gid < p->xid) break;
+    o = p;
+  }
+  /* uh oh, time to do a real lookup */
+  t = malloc(sizeof(struct xtable));
+  if ((ent = getgrgid(gid)) != NULL) t->name = scopy(ent->gr_name);
+  else {
+    sprintf(gbuf,"%d",gid);
+    t->name = scopy(gbuf);
+  }
+  t->xid = gid;
+  t->nxt = p;
+  if (o == gtable[gent]) gtable[gent] = t;
+  else o->nxt = t;
+  return t->name;
+}
+
+printit(s)
+char *s;
+{
+  char c;
+
+  if (Nflag) printf("%s",s);
+  else {
+    for(;*s;s++)
+      if (isprint(*s)) putchar(*s);
+      else {
+	if (qflag) putchar('?');
+	else {
+	  c = *s & 127;
+	  if (c < ' ') printf("^%c",c+'@');
+	  else if (c == 127) printf("^?");
+	}
+      }
+  }
+}
+
+
 /*
  * Hacked in DIR_COLORS support for linux. ------------------------------
  */
@@ -505,7 +661,7 @@ parse_dir_colors()
     return;
   }
 
-  term = SCOPY(s);
+  term = scopy(s);
 
   if ((s=getenv("LS_COLORS")) == NULL || (!force_color && (nocolor || !isatty(1)))) {
     colorize = FALSE;
@@ -516,7 +672,7 @@ parse_dir_colors()
 /*    ansilines = TRUE; */
   }
 
-  colors = SCOPY(s);
+  colors = scopy(s);
 
   arg = parse(colors);
 
@@ -524,37 +680,52 @@ parse_dir_colors()
     c = split(arg[i]);
     switch(cmd(c[0])) {
       case COL_NORMAL:
-	if (c[1]) norm_flgs = SCOPY(c[1]);
+	if (c[1]) norm_flgs = scopy(c[1]);
 	break;
       case COL_FILE:
-	if (c[1]) file_flgs = SCOPY(c[1]);
+	if (c[1]) file_flgs = scopy(c[1]);
 	break;
       case COL_DIR:
-	if (c[1]) dir_flgs = SCOPY(c[1]);
+	if (c[1]) dir_flgs = scopy(c[1]);
 	break;
       case COL_LINK:
-	if (c[1]) link_flgs = SCOPY(c[1]);
+	if (c[1]) link_flgs = scopy(c[1]);
 	break;
       case COL_FIFO:
-	if (c[1]) fifo_flgs = SCOPY(c[1]);
+	if (c[1]) fifo_flgs = scopy(c[1]);
 	break;
       case COL_SOCK:
-	if (c[1]) sock_flgs = SCOPY(c[1]);
+	if (c[1]) sock_flgs = scopy(c[1]);
 	break;
       case COL_BLK:
-	if (c[1]) block_flgs = SCOPY(c[1]);
+	if (c[1]) block_flgs = scopy(c[1]);
 	break;
       case COL_CHR:
-	if (c[1]) char_flgs = SCOPY(c[1]);
+	if (c[1]) char_flgs = scopy(c[1]);
 	break;
       case COL_EXEC:
-	if (c[1]) exec_flgs = SCOPY(c[1]);
+	if (c[1]) exec_flgs = scopy(c[1]);
+	break;
+      case COL_ORPHAN:
+	if (c[1]) orphan_flgs = scopy(c[1]);
+	break;
+      case COL_MISSING:
+	if (c[1]) missing_flgs = scopy(c[1]);
+	break;
+      case COL_LEFTCODE:
+	if (c[1]) leftcode = scopy(c[1]);
+	break;
+      case COL_RIGHTCODE:
+	if (c[1]) rightcode = scopy(c[1]);
+	break;
+      case COL_ENDCODE:
+	if (c[1]) endcode = scopy(c[1]);
 	break;
       case DOT_EXTENSION:
 	if (c[1]) {
 	  e = malloc(sizeof(struct extensions));
-	  e->ext = SCOPY(c[0]+1);
-	  e->flgs = SCOPY(c[1]);
+	  e->ext = scopy(c[0]+1);
+	  e->flgs = scopy(c[1]);
 	  if (ext == NULL) ext = e;
 	  else {
 	    e->nxt = ext;
@@ -566,7 +737,14 @@ parse_dir_colors()
   }
 
   /* make sure at least norm_flgs is defined.  We're going to assume vt100 support */
-  if (!norm_flgs) norm_flgs = SCOPY("00");
+  if (!leftcode) leftcode = scopy("\033[");
+  if (!rightcode) rightcode = scopy("m");
+  if (!norm_flgs) norm_flgs = scopy("00");
+
+  if (!endcode) {
+    sprintf(buf,"%s%s%s",leftcode,norm_flgs,rightcode);
+    endcode = scopy(buf);
+  }
 
 /*  if (!termmatch) colorize = FALSE; */
 }
@@ -610,7 +788,9 @@ char *s;
     char cmdnum;
   } cmds[] = {
     "no", COL_NORMAL, "fi", COL_FILE, "di", COL_DIR, "ln", COL_LINK, "pi", COL_FIFO,
-    "so", COL_SOCK, "bd", COL_BLK, "cd", COL_CHR, "ex", COL_EXEC, NULL, 0
+    "so", COL_SOCK, "bd", COL_BLK, "cd", COL_CHR, "ex", COL_EXEC, "mi", COL_MISSING,
+    "or", COL_ORPHAN, "lc", COL_LEFTCODE, "rc", COL_RIGHTCODE, "ec", COL_ENDCODE,
+    NULL, 0
   };
   int i;
 
@@ -620,49 +800,63 @@ char *s;
   return ERROR;
 }
 
-color(mode,name)
+color(mode,name,orphan,islink)
 u_short mode;
 char *name;
+char orphan, islink;
 {
   struct extensions *e;
   int l;
 
+  if (orphan) {
+    if (islink) {
+      if (missing_flgs) {
+	printf("%s%s%s",leftcode,missing_flgs,rightcode);
+	return TRUE;
+      }
+    } else {
+      if (orphan_flgs) {
+	printf("%s%s%s",leftcode,orphan_flgs,rightcode);
+	return TRUE;
+      }
+    }
+  }
   l = strlen(name);
   for(e=ext;e;e=e->nxt) {
     if (!strcmp(name+(l-strlen(e->ext)),e->ext)) {
-      printf("\033[%sm",e->flgs);
+      printf("%s%s%s",leftcode,e->flgs,rightcode);
       return TRUE;
     }
   }
   switch(mode & S_IFMT) {
     case S_IFIFO:
       if (!fifo_flgs) return FALSE;
-      printf("\033[%sm",fifo_flgs);
+      printf("%s%s%s",leftcode,fifo_flgs,rightcode);
       return TRUE;
     case S_IFCHR:
       if (!char_flgs) return FALSE;
-      printf("\033[%sm",char_flgs);
+      printf("%s%s%s",leftcode,char_flgs,rightcode);
       return TRUE;
     case S_IFDIR:
       if (!dir_flgs) return FALSE;
-      printf("\033[%sm",dir_flgs);
+      printf("%s%s%s",leftcode,dir_flgs,rightcode);
       return TRUE;
     case S_IFBLK:
       if (!block_flgs) return FALSE;
-      printf("\033[%sm",block_flgs);
+      printf("%s%s%s",leftcode,block_flgs,rightcode);
       return TRUE;
     case S_IFLNK:
       if (!link_flgs) return FALSE;
-      printf("\033[%sm",link_flgs);
+      printf("%s%s%s",leftcode,link_flgs,rightcode);
       return TRUE;
     case S_IFSOCK:
       if (!sock_flgs) return FALSE;
-      printf("\033[%sm",sock_flgs);
+      printf("%s%s%s",leftcode,sock_flgs,rightcode);
       return TRUE;
     case S_IFREG:
       if (!exec_flgs) return FALSE;
       if ((mode & S_IEXEC) | (mode & (S_IEXEC>>3)) | (mode & (S_IEXEC>>6))) {
-	printf("\033[%sm",exec_flgs);
+	printf("%s%s%s",leftcode,exec_flgs,rightcode);
 	return TRUE;
       }
       return FALSE;
